% -*- prolog -*-

% This file contains Praedicates used by mod_disconstr.pl and mod_equconstr.pl



% ----------------------------------------------
% get_summarizable_clauses(InputClause, 
%                          InputList, 
%                          SummarizableClauses, 
%                          RestOfInputList)
% ----------------------------------------------

% termination (InputList = [])
get_summarizable_clauses(_, [], [], []).

			 
% first element of InputList is a summarizable clause =>
% add this clause to SummarizableClauses
get_summarizable_clauses(InputClause,
                         InputList,
			 SummarizableClauses,
			 RestOfInputList) :-

	InputClause =.. [_, _, _, LiteralList1, _, _],
	InputList = [HeadOfInputList | TailOfInputList],
	HeadOfInputList =.. [_, U, V, LiteralList2, X, Y],

	% Do they have the same number of literals?
	same_length(LiteralList1, LiteralList2),
	
	% Do they have the same set of predicates?
	((cmd_arg(disconstr, (_, noperm));
	  cmd_arg(equconstr, (_, _, _, noperm))) ->
	        same_predicates_without_permutation(LiteralList1, 
	                                            LiteralList2),
		PermutedLiteralList2 = LiteralList2
	;
	        same_predicates(LiteralList1, 
	                        LiteralList2, 
				PermutedLiteralList2)
	),
	!, % No backtracking in this case
	
	% proceed recursively
	get_summarizable_clauses(InputClause,
	                         TailOfInputList,
				 TailOfSummarizableClauses,
				 RestOfInputList),

	SummarizableClauses = [clause(U, V, PermutedLiteralList2, X, Y) | 
                               TailOfSummarizableClauses].


% Anything else that's not summarizable (be it clause or rule)
% is simply appended to the OutputList
get_summarizable_clauses(Clause, 
                         [Entity | TailOfInputList],
			 SummarizableClauses, 
			 [Entity | TailOfRestOfInputList]) :-

	get_summarizable_clauses(Clause,
	                         TailOfInputList,
				 SummarizableClauses,
				 TailOfRestOfInputList).




% --------------------------------------------------------
% fact(LiteralList) - is the Literal positive?
% --------------------------------------------------------

fact([Literal]) :-

    Literal =.. [++, _].



% --------------------------------------------------------
% remove_subsumed_clauses(InputClauseList, OutputClauseList)
% - Generate OutputClauseList by removing subsumed clauses 
%   from InputClauseList. 
% --------------------------------------------------------

remove_subsumed_clauses([], []).


% If HeadClause is subsumed by a clause of TailOfInputClauseList,
% HeadClause is not a member of OutputClauseList.
% Check the TailOfInputClauseList.
remove_subsumed_clauses([HeadClause | TailOfInputClauseList], 
                        OutputClauseList) :-

	is_subsumed_by(HeadClause, 
                       TailOfInputClauseList),
	remove_subsumed_clauses(TailOfInputClauseList, 
                                OutputClauseList).


% Include HeadClause into OutputClauseList.
% If HeadClause subsumes clauses of TailOfInputClauseList,
% remove the subsumed clauses and check the rest of TailOfInputClauseList.
remove_subsumed_clauses([HeadClause | TailOfInputClauseList], 
                        [HeadClause | LocalOutputClauseList]) :-

	subsumes_clauses_of(HeadClause, 
	                    TailOfInputClauseList, 
			    RestOfTailOfInputClauseList),
	remove_subsumed_clauses(RestOfTailOfInputClauseList, 
                                LocalOutputClauseList).


% Include HeadClause into OutputClauseList.
% Check the TailOfInputClauseList.
remove_subsumed_clauses([HeadClause | TailOfInputClauseList], 
                        [HeadClause | LocalOutputClauseList]) :-

	remove_subsumed_clauses(TailOfInputClauseList, 
                                LocalOutputClauseList).



% --------------------------------------------------------
% is_subsumed_by(Clause, InputClauseList)
% - Clause is subsumed by InputClauseList if
%   1. the head clause of InputClauseList subsumes Clause or
%   2. Clause is subsumed by the tail of InputClauseList.
% --------------------------------------------------------

is_subsumed_by(_, []) :- fail.


is_subsumed_by(Clause,
	       [HeadOfInputClauseList|_]) :-

	subsumes(HeadOfInputClauseList, Clause).


is_subsumed_by(Clause, 
	       [_|TailOfInputClauseList]) :-

	is_subsumed_by(Clause, TailOfInputClauseList).



% --------------------------------------------------------
% subsumes_clauses_of(Clause, InputClauseList, OutputClauseList)
% - Does Clause subsume clauses of InputClauseList ?
%   If so, OutputClauseList is InputClauseList without the 
%   subsumed clauses.
% --------------------------------------------------------

% Termination: 
% If InputClauseList is empty, OutputClauseList is empty.
subsumes_clauses_of(_, [], []).


% If Clause subsumes HeadOfInputClauseList, 
% HeadOfInputClauseList is not a member of OutputClauseList.
subsumes_clauses_of(Clause, 
	            [HeadOfInputClauseList | TailOfInputClauseList], 
		    OutputClauseList) :-

	subsumes(Clause, HeadOfInputClauseList),
	subsumes_clauses_of(Clause, 
	                    TailOfInputClauseList, 
			    OutputClauseList).


% If Clause does not subsume HeadOfInputClauseList, 
% HeadOfInputClauseList is a member of OutputClauseList.
subsumes_clauses_of(Clause, 
	            [HeadOfInputClauseList | TailOfInputClauseList], 
		    [HeadOfInputClauseList | OutputClauseList]) :-

	subsumes_clauses_of(Clause, 
	                    TailOfInputClauseList, 
			    OutputClauseList).



% --------------------------------------------------------
% subsumes(Clause1, Clause2)
% - Does Clause1 subsume Clause2 ?
%   If so, 
%   1. both clauses are declarative
%   2. the literal list of Clause1 is not longer than the literal list 
%      of Clause2
%   3. the literal list of Clause2 contains a permutation of the 
%      literals of Clause2
%   4. the permuted and shortened (to the length of the literal list of 
%      Clause1) literal list of Clause2 is an instance of the literal 
%      list of Clause1
% --------------------------------------------------------
	
subsumes(Clause1, Clause2) :-

	Clause1 =.. [_, _, _, LiteralList1, _, _],
	Clause2 =.. [_, _, _, LiteralList2, _, _],

	LiteralList1 == LiteralList2.


	
subsumes(Clause1, Clause2) :-

	Clause1 =.. [_, _, _, LiteralList1, _, _],
	Clause2 =.. [_, _, _, LiteralList2, _, _],

	% LiteralList1 is not longer than LiteralList2
	length(LiteralList1, L1),
	length(LiteralList2, L2),
	L1 =< L2,

	% LiteralList2 contains a permutation of the literals of 
	% LiteralList1
	same_predicates(LiteralList1, 
                        LiteralList2, 
			PermutedLiteralList2),

	instance(PermutedLiteralList2, LiteralList1).



% ------------------------------------------------------------------
% same_predicates_without_permutation(LiteralList1, LiteralList2)
%
% - Does LiteralList2 contain all the literals of LiteralList1 ?
%   LiteralList2 has to be of the same length or longer than
%   LiteralList1.
% ------------------------------------------------------------------

same_predicates_without_permutation(LiteralList1, []) :- 
	LiteralList1 \== [],
	fail.

same_predicates_without_permutation([], _).

same_predicates_without_permutation([Literal1 | TailOfLiteralList1], 
	                            [Literal2 | TailOfLiteralList2]) :-

	equivalent_literals(Literal1, Literal2),
	same_predicates_without_permutation(TailOfLiteralList1, 
                                            TailOfLiteralList2).



% ------------------------------------------------------------------
% same_predicates(LiteralList1, LiteralList2, PermutedLiteralList2)
%
% - Does LiteralList2 contain all the literals of LiteralList1 ?
%   PermutedLiteralList2 contains only that literals from LiteralList2 
%   which appear in LiteralList1. These are in the same order as they 
%   appear in LiteralList1.
%   LiteralList2 has to be of the same length or longer than
%   LiteralList1.
% ------------------------------------------------------------------

same_predicates(LiteralList1, [], []) :- 
	LiteralList1 \== [],
	fail.

same_predicates([], _, []).

same_predicates([Literal1 | TailOfLiteralList1], LiteralList2, 
	        [Literal2 | LocalPermutedLiteralList2]) :-

	% try to find a literal (Literal2) in LiteralList2, 
	% that's equivalent to Literal1
	check_literal(Literal1, LiteralList2, Literal2, RestOfLiteralList2),

        % proceed recursively
	same_predicates(TailOfLiteralList1, 
                        RestOfLiteralList2, 
			LocalPermutedLiteralList2).


% ------------------------------------------------------------------
% check_literal(InputLiteral, InputLiteralList, 
%               OutputLiteral, RestInputLiteralList)
% try to find a literal (OutputLiteral) in InputLiteralList, 
% that's equivalent to InputLiteral
% RestInputLiteralList = InputLiteralList \ OutputLiteral
% ------------------------------------------------------------------

check_literal(_,[],_,_) :- fail.

check_literal(InputLiteral,
	      [HeadInputLiteralList | TailInputLiteralList],
              HeadInputLiteralList,
	      TailInputLiteralList) :-

	equivalent_literals(InputLiteral, HeadInputLiteralList), 
	!.

check_literal(InputLiteral,
              [HeadInputLiteralList | TailInputLiteralList],
              OutputLiteral,
              [HeadInputLiteralList | TailOutputLiteralList]) :-

	check_literal(InputLiteral,
                      TailInputLiteralList,
		      OutputLiteral,
		      TailOutputLiteralList).


% ------------------------------------------------------------------
% equivalent_literals(Literal1, Literal2)
% Two literals are equivalent, if 
% 1. they're either both positive or negative
% 2. they've got the same predicate-symbol
% 3. they're of the same arity
% ------------------------------------------------------------------

% The literals are identical. 
equivalent_literals(Literal1, Literal2) :-

	Literal1 == Literal2.


equivalent_literals(++ ~ Atom1, ++ ~ Atom2) :-

    !, % do not try the other cases

    % same predicate-symbol (Pred)
    Atom1 =.. [Pred | Args1],
    Atom2 =.. [Pred | Args2],

    % same arity
    same_length(Args1, Args2).


equivalent_literals(++ Atom1, ++ Atom2) :-

    !, % do not try the other cases

    % same predicate-symbol (Pred)
    Atom1 =.. [Pred | Args1],
    Atom2 =.. [Pred | Args2],

    % same arity
    same_length(Args1, Args2).


equivalent_literals(-- ~ Atom1, -- ~ Atom2) :-

    !, % do not try the other cases

    % same predicate-symbol (Pred)
    Atom1 =.. [Pred | Args1],
    Atom2 =.. [Pred | Args2],

    % same arity
    same_length(Args1, Args2).


equivalent_literals(-- Atom1, -- Atom2) :-

    !, % do not try the other cases

    % same predicate-symbol (Pred)
    Atom1 =.. [Pred | Args1],
    Atom2 =.. [Pred | Args2],

    % same arity
    same_length(Args1, Args2).


% ------------------------------------------------------------------
% get_list_of_dictionaries(ClauseList, AppendedDictionaries)
% Append the Dictionaries of all clasues in ClauseList.
% ------------------------------------------------------------------

append_dictionaries([],[]).

append_dictionaries([Clause | TailClauseList], AppendedDictionaries) :-

    Clause =.. [_,_,_,_,_,Dictionary],
    append_dictionaries(TailClauseList, LocalDictionaries),
    append(Dictionary, LocalDictionaries, AppendedDictionaries).


% ------------------------------------------------------------------
% get_list_of_arglists(SummarizableClauses, ListOfArgLists)
% Generate ListOfArgLists. Each of its elements is a list of the arguments
% appearing in a clause from SummarizableClauses.
% ------------------------------------------------------------------

get_list_of_arglists([],[]).

get_list_of_arglists([Clause | TailClauseList], 
	             [FirstArgList | TailArgLists]) :-

    Clause =.. [_,_,_, LiteralList,_,_],

    % generate FirstArgList, which contains the arguments of the 
    % literals in LiteralList (which is added to ListOfArgLists)
    get_arglist(LiteralList, FirstArgList),

    % proceed recursively
    get_list_of_arglists(TailClauseList, TailArgLists).





% ------------------------------------------------------------------
% get_arglist(LiteraLList, ArgList)
% Generate ArgList, which contains the arguments of the literals in LiteralList
% e.g. LiteralList = [++p(X, Y, Z), --q(f(a, Z)), ++r(g(f(a, X), Y), b)]
%      ArgList     = [    X, Y, Z,      f(a, Z),      g(f(a, X), Y), b ]
% ------------------------------------------------------------------

get_arglist([], []).

% Handle Literals with negation symbol ~
get_arglist([HeadLiteral | TailLiteralList], ArgList) :-

    % get the arguments of HeadLiteral
    HeadLiteral =.. [_, NegatedAtom],
    NegatedAtom =.. [~, Atom],
    Atom =.. [_ | Arguments],

    % recursively get the arglist of TailLiteralList 
    get_arglist(TailLiteralList, LocalArgList),
    append(Arguments, LocalArgList, ArgList).

% Handle Literals without negation symbol ~
get_arglist([HeadLiteral | TailLiteralList], ArgList) :-

    % get the arguments of HeadLiteral
    HeadLiteral =.. [_, Atom],
    \+ Atom =.. [~ | _],
    Atom =.. [_ | Arguments],

    % recursively get the arglist of TailLiteralList 
    get_arglist(TailLiteralList, LocalArgList),
    append(Arguments, LocalArgList, ArgList).



% ------------------------------------------------------------------
% gen_new_literallist(LiteralList, ListOfArgLists, 
%                     VarCounter, InVarList
%                     NewLiteralList, NewListOfArgLists, 
%                     OutVarList)
% Rename all of the n arguments in LiteralList to V1, ..., Vn 
% and return NewLiteralList and VarList = [V1, ..., Vn] 
% Thus the clause p(X,  Y,  Z)  v ~q(f(a, Z)) v r(g(f(a, X), Y), b) becomes
%                 p(V1, V2, V3) v ~q(V4)      v r(V5,            V6)
% ------------------------------------------------------------------

gen_new_literallist([], 
                    ListOfArgLists, 
	            _, 
		    InVarList, 
		    [], 
		    ListOfArgLists, 
		    InVarList).

% Handle Literals with negation symbol ~
gen_new_literallist([Literal | TailLiteralList],
                    ListOfArgLists,
		    Position, 
		    InVarList,
                    [NewLiteral | TailNewLiteralList], 
                    NewListOfArgLists,
                    OutVarList) :-

    Literal        =.. [Sign, NegatedAtom],
    NegatedAtom    =.. [~, Atom], 
    Atom           =.. [Praed | Arguments],

    % rename the arguments of Literal 
    process_arguments(Arguments, NewArguments, 
                      Position, NewPosition, 
                      ListOfArgLists, LocalNewListOfArgLists,
		      InVarList, LocalVarList),
    NewAtom        =.. [Praed | NewArguments],
    NewNegatedAtom =.. [~, NewAtom],
    NewLiteral     =.. [Sign, NewNegatedAtom], 

    % recursively process the remaining arguments
    gen_new_literallist(TailLiteralList, LocalNewListOfArgLists,
	                NewPosition, LocalVarList,
                        TailNewLiteralList, NewListOfArgLists,
			OutVarList).


% Handle Literals without negation symbol ~
gen_new_literallist([Literal | TailLiteralList], 
                    ListOfArgLists,
		    Position, 
		    InVarList,
                    [NewLiteral | TailNewLiteralList], 
                    NewListOfArgLists,
                    OutVarList) :-

    Literal    =.. [Sign, Atom],
    \+ Atom =.. [~ | _],
    Atom       =.. [Praed | Arguments],

    % rename the arguments of Literal 
    process_arguments(Arguments, NewArguments, 
                      Position, NewPosition,
                      ListOfArgLists, LocalNewListOfArgLists,
		      InVarList,LocalVarList),
    NewAtom    =.. [Praed | NewArguments],
    NewLiteral =.. [Sign, NewAtom], 

    % recursively process the remaining arguments
    gen_new_literallist(TailLiteralList, LocalNewListOfArgLists,
	                NewPosition, 
			LocalVarList,
                        TailNewLiteralList, NewListOfArgLists,
                        OutVarList).


% ------------------------------------------------------------------
% process_arguments(ArgList, NewArgList, 
%                   Position, OutPosition,
%                   ListOfArgLists, NewListOfArgLists,
%       	    InVarList, OutVarList) :-
% Rename all of the n arguments in ArgList to V1, ..., Vn 
% and check, if columns can be removed.
% Return the generated argument list NewArgList, 
% the new variable counter OutPosition, 
% the changed list of argument lists NewListOfArgLists, 
% and the generated variable list NewListOfArgLists.
% ------------------------------------------------------------------

% Handle empty LiteralList
process_arguments([], [], 
                  Position, Position, 
                  ListOfArgLists, ListOfArgLists,
		  VarList, VarList).


% If the Position-th column of the argument list matrix 
% is equal to the (Position - i)-th column (for i>0),
% then remove the Position-th column from the argument list matrix
% and insert V(Position - i) into the new argument list 
% (do not generate a new variable).
process_arguments([_|Arguments], [NewArgument|NewArguments], 
                  Position, OutPosition,
                  ListOfArgLists, NewListOfArgLists,
		  InVarList, OutVarList) :-

    remove_repeated_column(ListOfArgLists, 
			   Position,
			   NrOfEqualColumn,
			   LocalListOfArgLists),

    get_term(NrOfEqualColumn, InVarList, NewArgument),

    process_arguments(Arguments, NewArguments, 
                      Position, OutPosition,
                      LocalListOfArgLists, NewListOfArgLists,
		      InVarList, OutVarList).


% Compute a generalization of the elements of the Position-th column 
% of the argument list matrix and replace each elements of the 
% Position-th column by the respective generalizer.
process_arguments([_|Arguments], [Generalization|NewArguments], 
                  InPosition, OutPosition,
                  ListOfArgLists, NewListOfArgLists,
		  InVarList, OutVarList) :-

	copy_term(ListOfArgLists, CopiedListOfArgLists),
	get_list_of_generalizable_terms2(InPosition, 
	                                 CopiedListOfArgLists, 
					 ListOfGeneralizableTerms),
	generalize_termlists(ListOfGeneralizableTerms, 
	                     Generalization),

	antiunify_column(ListOfArgLists,
	                 InPosition,
	                 Generalization, 
			 LocalListOfArgLists1),

	term_variables(Generalization, GeneralizationVarList),
	
	% 
	(re_test_input_problem(Generalization, GeneralizationVarList) ->
	    writeln("# re-test input problem"),
	    flush(output)
	;
	    flush(output)
	),

	eliminate_repeated_columns(LocalListOfArgLists1,
	                           InPosition,
				   InVarList,
				   GeneralizationVarList,
				   LocalListOfArgLists2,
	                           LocalPosition,
				   LocalVarList),


	process_arguments(Arguments, NewArguments, 
                          LocalPosition, OutPosition,
			  LocalListOfArgLists2, NewListOfArgLists,
			  LocalVarList, OutVarList).


re_test_input_problem(Generalization, GeneralizationVarList) :-
	
	\+ var(Generalization),
	functor(Generalization, _, Arity),
	Arity > 0,

	length(GeneralizationVarList, L),
	L > 0.



remove_repeated_column(_, 1, _, _) :- fail.

remove_repeated_column([], _, _, []).


remove_repeated_column([FirstArgList | RemainingArgLists], 
		       ColumnNr,
		       NrOfEqualColumn,
		       [NewFirstArgList | LocalNewListOfArgLists]) :-

	find_equal_argument(FirstArgList, 
	                    ColumnNr,
			    NrOfEqualColumn),

	remove_arg_from_arglist(FirstArgList, ColumnNr, NewFirstArgList),

	remove_repeated_column(RemainingArgLists, 
	                       ColumnNr, 
			       NrOfEqualColumn, 
			       LocalNewListOfArgLists).
	


find_equal_argument(_, 1, _) :- fail.


find_equal_argument([FirstArg | RemainingArgs], ArgNr, 1) :-
	Counter is ArgNr - 1,
	equal_column_args(FirstArg, RemainingArgs, Counter).


find_equal_argument([_|RemainingArgs], ArgNr, NrOfEqualArg) :-

	NewArgNr is ArgNr - 1,

	find_equal_argument(RemainingArgs,
	                    NewArgNr,
			    LocalNrOfEqualArg),

	NrOfEqualArg is LocalNrOfEqualArg + 1.
	



equal_column_args(Arg, [FirstArg|_], 1) :-
	!,
	Arg == FirstArg.

equal_column_args(Arg, [_|RemainingArgs], Counter) :-
	NewCounter is Counter - 1,
	equal_column_args(Arg, RemainingArgs, NewCounter).



remove_arg_from_arglist(_, 0, _) :- fail.

remove_arg_from_arglist([_| RemainingArgs], 1, RemainingArgs).

remove_arg_from_arglist([FirstArg | RemainingArgs], 
                        Counter, 
			[FirstArg | NewArgList]) :-

	NewCounter is Counter - 1,
	remove_arg_from_arglist(RemainingArgs, NewCounter, NewArgList).




antiunify_column([], _, _, []).

antiunify_column(ListOfArgLists, Position, Generalization, NewListOfArgLists) :-

	ListOfArgLists = [ArgList | TailListOfArgLists],
	get_term(Position, ArgList, Arg),

	copy_term(Generalization, LocalGeneralization),
	term_variables(LocalGeneralization, LocalGeneralizationVarList),

	never_failing_unify(LocalGeneralization, Arg),
	replace_term_by_termlist(Position, 
	                         ArgList, 
				 LocalGeneralizationVarList, 
				 NewArgList),

	antiunify_column(TailListOfArgLists, 
	                 Position, 
			 Generalization, 
			 TailNewListOfArgLists),
	NewListOfArgLists = [NewArgList | TailNewListOfArgLists].



eliminate_repeated_columns(InListOfArgLists,
	                   InPosition,
			   InVarList,
			   [],
			   InListOfArgLists,
	                   InPosition,
			   InVarList).

eliminate_repeated_columns(InListOfArgLists,
	                   InPosition,
			   InVarList,
			   GeneralizationVarList,
			   OutListOfArgLists,
	                   OutPosition,
			   OutVarList) :-

	GeneralizationVarList = [HeadVar | TailGeneralizationVarList],

	(remove_repeated_column(InListOfArgLists, 
	                        InPosition,
				NrOfEqualColumn,
				LocalListOfArgLists) ->
		get_term(NrOfEqualColumn, InVarList, VarAtom),
		LocalPosition = InPosition,
		LocalVarList = InVarList
	;
		LocalListOfArgLists = InListOfArgLists,
		rename_varlist([HeadVar], [VarAtom]),
		LocalPosition is InPosition + 1,
		append(InVarList, [VarAtom], LocalVarList)
	),

	HeadVar = VarAtom,
	eliminate_repeated_columns(LocalListOfArgLists,
	                           LocalPosition,
				   LocalVarList,
				   TailGeneralizationVarList,
				   OutListOfArgLists,
				   OutPosition,
				   OutVarList).



make_elem_constrs_triv_sat([], [], [], []).
	  

make_elem_constrs_triv_sat(VarList, _, _, _) :-
	VarList  = [Var | _],
	\+ (atom(Var), name(Var, [86 | _])),  % Var is not an atom Vi
	\+ (atom(Var), name(Var, [88 | _])),  % Var is not an atom Xi
	!,
	writeln('ERROR: Varlist contains wrong element: '),
	write(Var),
	writeln(' '),
	flush(output),
	fail.


make_elem_constrs_triv_sat(VarList, 
	                   TermList, 
			   OutVarList, 
			   OutTermList) :-
	VarList  = [Var  | TailVarList],
	TermList = [Term | TailTermList],
	(var(Term) ->
	    % unify Var with Term and remove the elem constr [Var] = [Term]
	    Var = Term,
	    OutVarList  = LocalTailVarList,
	    OutTermList = LocalTailTermList
	;  
	    % do not change the elem constr [Var] = [Term]
	    OutVarList  = [Var  | LocalTailVarList],
	    OutTermList = [Term | LocalTailTermList]
	),
	make_elem_constrs_triv_sat(TailVarList, 
	                           TailTermList,
				   LocalTailVarList,
				   LocalTailTermList).


% ------------------------------------------------------------------
% utilities
% ------------------------------------------------------------------

same_length(List1, List2) :-
    length(List1, N),
    length(List2, N).




rename_variables([], []).

% Prefix the output names of all variables
rename_variables([LeftSide = RightSide|TailOfDict], 
	         [NewLeftSide = RightSide|NewTailOfDict]) :-

        % convert LeftSide to a list of ascii-values
	name(LeftSide, AsciiList),

	% 95 => "_"
	name(NewLeftSide, [95 | AsciiList]),

	rename_variables(TailOfDict, NewTailOfDict).



rename_varlist([], []).

rename_varlist([_| TailVarlist], [NewFirstVar | TailNewVarlist]) :-

	getval(rename_counter, Number),
	NewNumber is Number + 1,
	setval(rename_counter, NewNumber),
	
        % convert Number to a list of ascii-values
	name(Number, VarNameAsciiList),

	% 88 => "X"
	name(NewFirstVar, [88 | VarNameAsciiList]),

	rename_varlist(TailVarlist, TailNewVarlist).





% Write list without abbreviations of the prolog system:
write_list([]) :-
	writeln('\n'),
	flush(output).

write_list([Head|Tail]) :-
	writeln(Head),
	write_list(Tail),
	flush(output).



copy_clauses([], []).

copy_clauses([HeadClause|TailOfClauseList],
	     [NewHeadClause|NewTailOfClauseList]) :-

	copy_term(HeadClause, NewHeadClause),
	copy_clauses(TailOfClauseList, NewTailOfClauseList).

