% -*- prolog -*-

% -----------------------------------------------------------
% disconstr/4
% generates constraints of the form "$disconstr([[[...],[...]],...], N)" 
% which are processed by the built-in defined in sam/i_disconstr.c
% disconstr/4 always succeeds
% ----------------------------------------------------------- 

disconstr(problemwise, [], _, []).

disconstr(problemwise, InputClauseList, _, OutputClauseList) :-

    copy_clauses(InputClauseList, NewInputClauseList),
    remove_subsumed_clauses(NewInputClauseList, RestOfInputClauseList),
    setval(rename_counter, 1),
    gen_disconstrs(RestOfInputClauseList, OutputClauseList).


% -----------------------------------------------------------
% process_switch/1
% -----------------------------------------------------------

process_switch_disconstr( ["-disconstr" | ArgList] ) :-
	
	(occurs(onlyfacts, ArgList) ->
	    Argument1 = onlyfacts
	;
	    Argument1 = []
	),
	
	(occurs(noperm, ArgList) ->
	    Argument2 = noperm
	;
	    Argument2 = []
	),
	
	assert( cmd_arg(disconstr, (Argument1, 
	                            Argument2)) ),
	write("# *** option:  (disconstr"),
	write(Argument1),
	write(", "),
	write(Argument2),
	writeln(")"),
	flush(output).



% -----------------------------------------------------------
% gen_disconstrs([Clause | InputClauseList], 
%                [NewClause|LocalOutputClauseList])
% gen_disconstrs/2 finds all clauses in InputClauseList, 
% which together with Clause can be summarized as the single 
% clause NewClause.
% NewClause is added to the output clause list (2nd arg).
% This is done recursively for all clauses in InputClauseList.
% ------------------------------------------------------------  

gen_disconstrs([], []). 


gen_disconstrs([Clause | InputClauseList], [NewClause|LocalOutputClauseList]) :-

    % find all the clauses suitable for the summarization 
    % and put them into ListOfSummarizableClauses.
    % RestOfInputClauseList is the rest of InputClauseList,
    % i.e. InputClauseList \ ListOfSummarizableClauses
    get_summarizable_entities_disconstr(Clause, 
			                InputClauseList,
					ListOfSummarizableClauses,
					RestOfInputClauseList),

    % complete the summarization into NewClause
    summarize_clauses_disconstr([Clause | ListOfSummarizableClauses],
		                NewClause),
    
    % proceed recursively
    gen_disconstrs(RestOfInputClauseList, 
		   LocalOutputClauseList).



% ------------------------------------------------------------------
% get_summarizable_entities_disconstr(Entity, 
%                                     InputList, 
%                                     SummarizableEntities, 
%                                     RestOfInputList) 
% finds all the entities (facts and/or clauses) in InputList which 
% can be summarized with Entity and writes them into SummarizableEntities.
% RestOfInputList = InputList \ SummarizableEntities
%
% Two entities are summarizable, if 
% 1. they've got the same number of literals
% 2. they've got the same set of predicates (same_predicates/3)
% -------------------------------------------------------------------

% The entity is a clause. Only try to summarize if the onlyfacts option
% cmd_arg(disconstr, (onlyfacts, _)) is NOT used
get_summarizable_entities_disconstr(InputClause, 
                                    InputList, 
				    SummarizableEntities, 
				    RestOfInputList) :-

	cmd_arg(disconstr, ([], _)), % NOT cmd_arg(disconstr, (onlyfacts, _))
	get_summarizable_clauses(InputClause, 
                                 InputList, 
				 SummarizableEntities, 
				 RestOfInputList).


get_summarizable_entities_disconstr(InputClause, 
                                    InputList, 
				    SummarizableEntities, 
				    RestOfInputList) :-

	cmd_arg(disconstr, (onlyfacts, _)),
	InputClause = [_, _, _, LiteralList, _, _],
	fact(LiteralList),
	get_summarizable_clauses(InputClause, 
	                         InputList, 
				 SummarizableEntities, 
				 RestOfInputList).


% Otherwise "return" empty SummarizableEntities (entity is either 
% a rule or not a fact & the onlyfacts-mode is selected)
get_summarizable_entities_disconstr(_, InputList, [], InputList).

	      

% ------------------------------------------------------------------
% summarize_clauses_disconstr(SummarizableClauses, OutputClause)
% Summarize SummarizableClauses into OutputClause, 
% which may contain a Literal (actually a constraint) "--$disconstr(...)"
% ------------------------------------------------------------------

summarize_clauses_disconstr([OutputClause],OutputClause).

summarize_clauses_disconstr(SummarizableClauses, OutputClause) :-

    % get list of arglists and append dictionaries
    % of the clauses in SummarizableClauses
    get_list_of_arglists(SummarizableClauses, ListOfArgLists),
    append_dictionaries(SummarizableClauses, AppendedDictionaries),

    % prefix the output names of all variables in order to avoid 
    % name-clashes with new generated variables Vi or Xi
    rename_variables(AppendedDictionaries, RenamedAppendedDictionaries),

    % with the help of this list, generate OutputClause from HeadClause
    SummarizableClauses = [HeadClause |_],
    gen_new_clause_disconstr(HeadClause, 
                             ListOfArgLists, 
			     RenamedAppendedDictionaries, 
			     OutputClause).
	


% ------------------------------------------------------------------
% gen_new_clause_disconstr(ExampleClause, ListOfArgLists, NewClause)
% Generates NewClause from ExampleClause and ListOfArgLists by renaming 
% all of the arguments and possibly adding a constraint:
% NewClause will contain a constraint ($disconstr), if ListOfArgLists 
% isn't empty.
% ------------------------------------------------------------------

% ListOfArgLists is empty
gen_new_clause_disconstr(ExampleClause, [], _, ExampleClause).


% ListOfArgLists isn't empty
gen_new_clause_disconstr(InClause, 
                         ListOfArgLists, 
			 AppendedDictionaries, 
			 OutClause) :- 

    InClause =.. [U, V, W, LiteralList, X, _],

    % rename all of the n arguments in LiteralList to V1, ..., Vn 
    % and return NewLiteralList and VarList = [V1, ..., Vn] 
    gen_new_literallist(LiteralList, ListOfArgLists, 
	                1, [],
                        NewLiteralList, NewListOfArgLists, 
                        VarList),

    % generate the constraint from ListOfArgLists and VarList
    gen_dis_constraint(NewListOfArgLists, VarList, EqConstraint),

    % finally generate OutClause
    length(ListOfArgLists, NrOfSummarizedClauses),
    OutClause =.. [U, V, W, 
                   [--'$disconstr'(EqConstraint, NrOfSummarizedClauses) | NewLiteralList], 
		   X, AppendedDictionaries].
    



% ------------------------------------------------------------------
% gen_dis_constraint(ListOfArgLists, VarList, Constraint)
% generate Constraint from  ListOfArglists and VarList
% ListOfArgLists = [ ArgList_1,            ArgList_2,           ...,  ArgList_n]
% Constraint     = [[VarList, ArgList_1], [VarList, ArgList_2], ..., [VarList, ArgList_n]]
% ------------------------------------------------------------------

gen_dis_constraint([], _, []).

gen_dis_constraint([ArgList | TailOfArgList], 
	           VarList, 
		   [[NewVarList, NewArgList] | TailOfConstraintList]) :-

	make_elem_constrs_triv_sat(VarList,
	                           ArgList,
	                           NewVarList,
				   NewArgList),

   gen_dis_constraint(TailOfArgList, VarList, TailOfConstraintList).
